STORED = b'STORED'
NOT_STORED = b'NOT_STORED'
TOUCHED = b'TOUCHED'
NOT_FOUND = b'NOT_FOUND'
DELETED = b'DELETED'
VERSION = b'VERSION'
OK = b'OK'
FLAG_PICKLE = 1 << 0
FLAG_INTEGER = 1 << 1
FLAG_JSON = 1 << 2
FLAG_BOOLEAN = 1 << 3
FLAG_STRING = 1 << 4
SERVER_RETRIES = 5
SOCKET_TIMEOUT = 3
DEAD_RETRY = 3
